
#########################################
# All the pairs of small/medium coronas #
#########################################

# all the small coronas
pc=small_coronas()

# all the medium coronas for each of then 10 lower bounds on s/r
#mcs=[medium_coronas(k) for k in pc[0:10]] # 5min 22s
mcs=load('mcs')

# compute all the medium corona equations, "just for fun"
# z=[equation(i,1) for i in mcs[1]] : 2h 21min 47s
# save(z,'eqmcs') # 36,2 Mo

# all the potential pairs small/medium coronas
pcmc=[mcs[pc.index(angles(sequence(k).replace('1','r')))] for k in pc]
# counts the 16805 pairs
sum([len(i) for i in pcmc])

#################################################
# Small and medium coronas assumed to intersect #
#################################################

# we shall assume that the small corona contains an r and the medium one contains an s
pcmc=[[l for l in pcmc[i] if pc[i][1]+pc[i][3]+pc[i][4]>0 and l[2]+l[4]+l[5]>0] for i in range(55)]
# 12265 pairs

#########################
# (pre)cover conditions #
#########################

# if xs appears in the vector angle l of the medium corona, then xr must appear in the vector angle k of the small corona; for x=1,r,s
precover=lambda (k,l,t): (l[2]==0 or k[1]>0) and (l[4]==0 or k[3]>0) and (l[5]==0 or k[4]>0)

# check the pre-cover condition
pcmc=[[l for l in pcmc[i] if precover((pc[i],l,1))] for i in range(55)]
# 2889

# if xsy appear in any medium corona of vector angle l, then xry or yrx must appear in the (unique) small corona of vector angle k
def cover(k,l,t):
    pseq=sequence(k)
    pseq=pseq+pseq[0:2] # avoids circularity problem
    voisins=[('1','1'),('1','r'),('1','s'),('r','r'),('r','s'),('s','s')]
    pxty=[(x+'r'+y in pseq or  y+'r'+x in pseq) for (x,y) in voisins]
    for tseq in sequences(l):
        tseq=tseq+tseq[0:2] # avoids circularity problem
        txsy=[(x+'s'+y in tseq or  y+'s'+x in tseq) for (x,y) in voisins]
        if min(vector(pxty)-vector(txsy))>=0:
            return true
    return false
    
# check the cover condition
pcmc=[[l for l in pcmc[i] if cover(pc[i],l,1)] for i in range(55)]
# 803 pairs in 1min 20s



##########################
# Hidden variable method #
##########################

# list of all the different r-coronas which appear in all the pairs small/medium coronas
mc=list(set(flatten(pcmc,max_level=1)))
# 192 r-coronas

# compute the equations associated with small and medium coronas
eqpc=[equation(i,0) for i in pc] # 6.5s
eqmc=[equation(i,1) for i in mc] # 1min 51s

# for each pair, compute the indices in lists pc and mc, order this
I=flatten([[(i,mc.index(j)) for j in pcmc[i]] for i in range(len(pc))],max_level=1)
I.sort()

# input: indices i and j of a small and a medium coronas,  k=AA or CIF
# output: cartesian product of the entries of the solutions of the associated equations
def projections((i,j),k):
    # compute the resultants
    R=ZZ[r](eqpc[i].resultant(eqmc[j],s))
    S=ZZ[s](eqpc[i].resultant(eqmc[j],r))
    RR=flatten([f[0].roots(k,multiplicities=false) for f in factor(R) if not useless(f[0])])
    SS=flatten([f[0].roots(k,multiplicities=false) for f in factor(S) if not useless(f[0])])
    # remove likely non real roots if computation on CIF
    if k==CIF:
        RR=[x.real() for x in RR if x.imag().contains_zero()]
        SS=[x.real() for x in SS if x.imag().contains_zero()]
    # filter real roots in (0,1)
    RR=[x for x in RR if x>0 and x<1]
    SS=[x for x in SS if x>0 and x<1]
    # cartesian product with s<r
    return flatten([[((i,j),(rr,ss)) for rr in RR if RIF(rr).endpoints()[1]>RIF(ss).endpoints()[0] and rr>ss] for ss in SS],max_level=1)

# applies projections on the whole I, result in RS or in E if error (see below)
def candidates(I,k):
    RS=[]
    E=[] # sometimes an error happens (see dedicated section below)
    for z in I:
        print(z)
        try:
            RS+=projections(z,k)
        except:
            E.append(z)
            print(sys.exc_info())
    return (RS,E)  

(RS,E)=candidates(I,AA)
# 56968 candidates in 4min
# 23 errors, see error section below


#################################
# arithmetic interval filtering #
#################################

# check that the small and medium coronas are compatibles with the interval arithmetic values of r and s
# check also that there is at least one large corona compatible with these values
def filter_int(RS):
    RS2=[]
    for ((i,j),(rr,ss)) in RS:
        print(((i,j),(rr,ss)))
        if not check_corona_int(pc[i],0,rr,ss):
            continue
        print('Values of r and s likely compatible with the small corona %s'%sequence(pc[i]))
        if not check_corona_int(mc[j],1,rr,ss):
            continue
        print('Values of r and s likely compatible with the medium corona %s'%sequence(mc[j]))
        if find_coronas_int(rr,ss,2)==[]:
            continue
        print('Values of r and s likely compatible with at least one large corona')
        RS2.append(((i,j),(rr,ss)))
    return RS2

RS2=filter_int(RS)
# 202 remaining candidates in 3min
# we can list those which have likely more than one small corona:
# however we cannot remove them without exact verification (after).

##########
# Errors #
##########

# several errors occur when computing the roots of an univariate polynomial on AA:

# 15 cases of ZeroDivisionError: (0, 179), (1, 57), (2, 9),  (2, 57),  (2, 190),  (3, 48), (3, 97), (4, 51), (4, 115), (5, 83), (7, 27), (8, 56), (8, 95), (22, 54), (39, 18)
# cause: the system is not zero dimensional, the resultant R or S is zero
# all pairs of coronas contain only r and s (no 1) except the two last ones (22, 54), (39, 18)
# two last ones: ruled out in the paper
# 13 first ones: must exists an s or r-corona with a 1 (otherwise no packing with 3 discs)
D1=[(0, 179), (1, 57), (2, 9),  (2, 57),  (2, 190),  (3, 48), (3, 97), (4, 51), (4, 115), (5, 83), (7, 27), (8, 56), (8, 95)]
# if exists an s-corona with 1 and r or an r-corona with 1 and s : case already handled otherwhere
# otherwise: s-corona without r or r-corona without s. check no solution:
# if there is an s-corona in 1 and s:
# equations in s for all the possible cases
eqs_ken_s=[equation(angles(u),0) for u in ['11111','1111s','111ss','11s1s','1111','11sss','1s1ss','111s','111','11ss']]
# check that with the two initial corona there is no solution
[[ideal([eqpc[i],eqmc[j],k]).variety() for k in eqs_ken_s] for (i,j) in D1]
# no solution in 2s
# if there is an r-corona in 1 and r: similar with r instead of s
eqs_ken_r=[i.substitute(s=r) for i in eqs_ken_s]
[[ideal([eqpc[i],eqmc[j],k]).variety() for k in eqs_ken_r] for (i,j) in D1]
# no solution in 8s

# RuntimeError: (25, 8)
# cause: maximum recursion depth exceeded while calling a Python object
# ValueError: (38, 8), (38, 76)
# cause: Refining interval that does not bound unique root
# AssertionError: (28, 76), (37, 11), (37, 80), (37, 111), (49, 76)
# cause: ? change the base ring of the polynomial from ZZ to AA (Thierry) ?
# all these errors:
E=[(25, 8),(28, 76), (37, 11), (37, 80), (37, 111), (49, 76), (38, 8), (38, 76)]
# solution: solve on CIF instead of AA, remove solutions whose imaginary part does not contain zero, set to zero the remaining imaginary part (i.e. we don't take any risk and keep these candidates)
(RSE,E2)=candidates(E,CIF)
# 5872 candidates in 1min, no more error

RSE2=filter_int(RSE)
# 1 remaining candidates in 1.2s : (28,76)=1rr1s/11rrs,
# ruled out combinatorially by lemma 6 -> NO MORE ERROR
# remark: exact solving via Groebner basis: 45minutes to show that it is ok without lemma 6


# dim 1 : doit y'avoir une s ou r corona avec 1
# si y'en a une avec contact s-r : cas déjà vu
# sinon : s-couronne en s et 1 ou r-couronne en r et 1 -> 10 valeurs Kennedy pour s ou r.
# essayer

##############################
# exact arithmetic filtering #
##############################

# check that there is exactly one small corona
# compute the medium coronas and avoid duplicates
# compute the large coronas and check that there is at least one
# return a list of elements ((i,j),(rr,ss),(p,m,g)) where (p,m,g) are all the small/medium/large coronas compatible with the small corona pc[i] and the medium mc[j], i.e., the value rr and ss of r and s
def filter_exact(RS):
    RS2=[]
    dup=[]  # to avoid duplicates
    k=0
    du=0
    for ((i,j),(rr,ss)) in RS:
        print(k)
        k+=1
        print(((i,j),(rr,ss)))
        # check that pc[i] is the only small corona
        p=[z for z in find_coronas_int(rr,ss,0) if check_corona_exact(z,0,rr,ss)]
        if p<>[pc[i]]:
            continue
        print('Only one small corona')
        # check that there is no previous small corona with the same set of medium coronas
        m=[z for z in find_coronas_int(rr,ss,1) if check_corona_exact(z,1,rr,ss)]
        if (i,m) in dup: # another case whose medium corona is in m
            du+=1
            continue
        print('%d medium coronas'%len(m))
        dup.append((i,m))
        # check that there is at least one large corona (this is the most time-consuming)
        g=[z for z in find_coronas_int(rr,ss,2) if check_corona_exact(z,2,rr,ss)]
        if g==[]:
            continue
        print('%d large coronas'%len(g))
        RS2.append(((i,j),(rr,ss),(p,m,g)))
    print('%d duplicates removed'%du)
    return RS2

RS3=filter_exact(RS2)
# 175 in 30min (mostly exact checking of large coronas, e.g. 39,130)
# removed from RS2: 26 duplicates and (11,157), which also allows s-corona 1s1ss (12): this is N°19, the only packing with two small coronas

# one could fasten by removing coronas with not well-filled holes (intersitital cases)
# it also helps for finding packing then...

#####################################
# Use Lemmas to rule out some pairs #
#####################################

# lemma 4
RS4=[i for i in RS3 if i[0][0] not in [10,13,17,37]]
# 151

# Lemma 5: 7 particular cases are removed
lem5=[
('1rsrs', '1rr1ss'),
('11rr', '11rrs'),
('1rr1s', '11rrs'),
('rrrrr', '1rsrsr'),
('rrrrs', '11rssr'),
('rrrss', '11rssr'),
('rrrs', '11rssr')]

RS5=[i for i in RS4 if (sequence(pc[i[0][0]]),sequence(mc[i[0][1]])) not in lem5]
# 145 left

# all: 48 minutes
